import configparser
import json
import logging
import os
from typing import Any, Iterable, Literal, Optional, TypeVar, cast

from pi_conf.module_check import has_yaml, is_tomllib
from pi_conf.provenance import Provenance, ProvenanceOp
from pi_conf.provenance import get_provenance_manager as get_pmanager

if has_yaml:
    import yaml
if is_tomllib:
    import tomllib
else:
    import toml


try:
    from platformdirs import site_config_dir
except:

    def site_config_dir(
        appname: str | None = None,
        appauthor: str | None | Literal[False] = None,
        version: str | None = None,
        multipath: bool = False,  # noqa: FBT001, FBT002
        ensure_exists: bool = False,  # noqa: FBT001, FBT002
    ) -> str:
        return f"~/.config/{appname}"


T = TypeVar("T", bound="AttrDict")

log = logging.getLogger(__name__)

sentinel = object()
_attr_dict_dont_overwrite = set([func for func in dir(dict) if getattr(dict, func)])


def _is_iterable_with_type(obj):
    try:
        if isinstance(obj, str):
            return False, None
        elif isinstance(obj, dict):
            return True, dict
        elif isinstance(obj, list):
            return True, list
        iter(obj)
        return True, None
    except TypeError:
        return False, None


class AttrDict(dict):
    """A dictionary class that allows referencing by attribute
    Example:
        d = AttrDict({"a":1, "b":{"c":3}})
        d.a.b.c == d["a"]["b"]["c"] # True
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.__dict__ = self

    def __post_init__(self):
        ## Iterate over members and add them to the dict
        members = [
            attr
            for attr in dir(self)
            if not callable(getattr(self, attr)) and not attr.startswith("__")
        ]
        for m in members:
            if isinstance(getattr(self, m), dict):
                nd = AttrDict.from_dict(getattr(self, m))
                setattr(self, m, nd)
                self[m] = nd
            else:
                self[m] = getattr(self, m)

    def __getattribute__(self, name: str) -> Any:
        """Get an attribute from the dictionary.
        This will allow you to access the dictionary keys as attributes.
        Returning Any removes MyPy errors."""
        return super().__getattribute__(name)

    def update(self, *args, **kwargs):
        """Update the config with another dict"""
        if "_no_attrdict" in kwargs:
            kwargs.pop("_no_attrdict")
            super().update(*args, **kwargs)
            return

        super().update(*args, **kwargs)
        AttrDict._from_dict(self, _depth=0, inline=True)

    def get_nested(
        self,
        keys: str,
        default: Any = sentinel,
        list_item: Optional[int] = 0,
        split_delimiter: str = ".",
    ) -> Any:
        """
        Get a nested value from the dictionary. Only works with string keys
        Args:
            keys (str): The keys to get from the dictionary
            default (Any): The default value if the key is not found
            list_item (int): If the key is a list, get the item at the index,
                set to None to disable
        Returns:
            Any: The value of the key

        Example:
            d = AttrDict({"a":1, "b":{"c":3}})
            d.get_nested('a') == 1
            d.get_nested('b.c') == 3
            d.get_nested('b.d', 'default') == 'default'
            d.get_nested('x.y.z', None) == None
            d.get_nested("notfound") # raises KeyError
        """
        current = self
        for key in keys.split(split_delimiter):
            if isinstance(current, dict):
                if key in current:
                    current = current[key]
                elif default is not sentinel:
                    return default
                else:
                    raise KeyError(f"Key not found: '{key}'")
            elif list_item is not None and isinstance(current, list):
                if list_item < len(current):
                    current = current[list_item]
                    if isinstance(current, dict) and key in current:
                        current = current[key]
                    elif default is not sentinel:
                        return default
                    else:
                        raise KeyError(f"Key not found: '{key}'")
                else:
                    if default is not sentinel:
                        return default
                    else:
                        raise KeyError(
                            f"'{key}', List index out of range: idx={list_item}, len={len(current)}"
                        )
            elif default is not sentinel:
                return default
            else:
                raise KeyError(f"'{key}' is not a nested dictionary")
        return current

    def to_env(
        self,
        recursive: bool = True,
        to_upper: bool = True,
        overwrite: bool = False,
        ignore_complications: bool = True,
    ) -> list[str | Any]:
        """recursively export the config to environment variables
        with the keys as prefixes
        args:
            d (dict): The dictionary to convert to an AttrDict
            recursive (bool): If True, recursively convert the dictionary to environment variables
            to_upper (bool): If True, convert the keys to uppercase
            overwrite (bool): If True, overwrite existing environment variables
            ignore_complications (bool): If True, ignore any complications in the dictionary
        returns:
            list: A list of tuples of the environment variables added
        """
        return self._to_env(
            recursive=recursive,
            to_upper=to_upper,
            overwrite=overwrite,
            ignore_complications=ignore_complications,
        )

    def _to_env(
        self,
        d: Optional[str | list[Any] | dict[str, Any] | Iterable] = None,
        recursive: bool = True,
        to_upper: bool = True,
        overwrite: bool = False,
        ignore_complications: bool = True,
        prefix: str = "",
        path: Optional[list] = None,
    ) -> list[str | Any]:
        """recursively export the config to environment variables
        with the keys as prefixes
        """
        if not path:
            path = []
        added_envs: list[str | Any] = []
        if d is None:
            d = self
        is_iterable, iterable_type = _is_iterable_with_type(d)
        if not is_iterable:
            v = json.dumps(d) if not isinstance(d, str) else d

            newk = "_".join(path)
            if to_upper:
                newk = newk.upper()
            if not os.environ.get(newk) or overwrite:
                os.environ[newk] = v
                added_envs.append((newk, v))
        elif iterable_type == dict:
            d = cast(dict[str, Any], d)
            for k, v in d.items():
                np = path.copy()
                np.append(k)
                a = self._to_env(
                    d=v,
                    recursive=recursive,
                    to_upper=to_upper,
                    overwrite=overwrite,
                    ignore_complications=ignore_complications,
                    prefix=f"{prefix}{k}",
                    path=np,
                )
                added_envs.extend(a)
        elif iterable_type == list:
            for i, v in enumerate(d):
                np = path[:-1].copy()
                np.append(f"{prefix}{i}")
                a = self._to_env(
                    d=v,
                    recursive=recursive,
                    to_upper=to_upper,
                    overwrite=overwrite,
                    ignore_complications=ignore_complications,
                    path=np,
                )
                added_envs.extend(a)
        elif not ignore_complications and is_iterable:
            raise Exception(f"Error! Cannot export iterable to environment variable d={d}")

        return added_envs

    @classmethod
    def _from_dict(
        cls: type[T],
        d: dict,
        _nested_same_class: bool = False,
        _depth: int = 0,
        inline: bool = False,
    ) -> T:
        """Make an AttrDict object without any keys
        that will overwrite the normal functions of a dict

        Args:
            cls (AttrDict): Create a new AttrDict object (or subclass)
            d (dict): The dictionary to convert to an AttrDict
            _nested_same_class (bool): If True, nested dicts will be the subclass,
                else they will be AttrDict

        Returns:
            AttrDict: the AttrDict object, or subclass
        """
        if not (_nested_same_class or _depth == 0):
            cls = AttrDict

        def _from_list_or_tuple(l):
            ### TODO change to generic iterable
            new_l = []
            for pot_dict in l:
                if isinstance(pot_dict, dict):
                    new_l.append(
                        cls._from_dict(
                            pot_dict, _nested_same_class=_nested_same_class, _depth=_depth + 1
                        )
                    )
                elif isinstance(pot_dict, list) or isinstance(pot_dict, tuple):
                    new_l.append(_from_list_or_tuple(pot_dict))
                else:
                    new_l.append(pot_dict)
            return new_l

        if not inline:
            d = cls(**d)
        for k, v in d.items():
            if k in _attr_dict_dont_overwrite:
                raise Exception(f"Error! config key={k} would overwrite a default dict attr/func")
            if isinstance(v, dict):
                d[k] = cls._from_dict(v, _nested_same_class=_nested_same_class, _depth=_depth + 1)
            elif isinstance(v, list) or isinstance(v, tuple):
                d[k] = _from_list_or_tuple(v)
            else:
                d[k] = v
        return d

    @classmethod
    def from_dict(
        cls: type["AttrDict"],
        d: dict,
        _nested_same_class: bool = False,
    ) -> "AttrDict":
        """Make an AttrDict object without any keys
        that will overwrite the normal functions of a

        Args:
            cls (AttrDict): Create a new AttrDict object (or subclass)
            d (dict): The dictionary to convert to an AttrDict
            _nested_same_class (bool): If True, nested dicts will be the subclass,
                else they will be AttrDict

        Returns:
            AttrDict: the AttrDict object, or subclass
        """
        return cls._from_dict(d, _nested_same_class=_nested_same_class, _depth=0)

    @classmethod
    def from_str(
        cls: type["AttrDict"],
        config_str: str,
        config_type: str = "toml",
        _nested_same_class: bool = False,
    ) -> "AttrDict":
        """Make an AttrDict object from a string

        Args:
            cls (AttrDict): Create a new AttrDict object (or subclass)
            config_str (str): The string to convert to an AttrDict
            config_type (str): The type of string to convert from (toml|json|ini|yaml)
            _nested_same_class (bool): If True, nested dicts will be the subclass,
                else they will be AttrDict

        Raises:
            Exception: _description_

        Returns:
            AttrDict: the AttrDict object, or subclass
        """
        if config_type == "toml":
            if is_tomllib:
                d = tomllib.loads(config_str)  # type: ignore
            else:
                d = toml.loads(config_str)  # type: ignore
        elif config_type == "json":
            d = json.loads(config_str)
        elif config_type == "ini":
            cfg_parser = configparser.ConfigParser()
            cfg_parser.read_string(config_str)
            d = {}
            for section in cfg_parser.sections():
                d[section] = {}
                for k, v in cfg_parser.items(section):
                    d[section][k] = v
        elif config_type == "yaml":
            if not has_yaml:
                raise Exception(
                    "Error! YAML not installed. If you would like to use YAML with pi-conf, "
                    'install it with `pip install pyyaml` or `pip install "pi-conf[yaml]"`'
                )
            d = yaml.safe_load(config_str)  # type: ignore
        else:
            raise Exception(f"Error! Unknown config_type '{config_type}'")
        return cls.from_dict(d, _nested_same_class=_nested_same_class)



""" Custom BaseSettings class for loading in complex types from toml files 
using the Config class."""

import json
import tempfile
from typing import Any, Dict, List, Self, Type, get_args, get_origin

import toml
from pydantic import BaseModel, ValidationError
from pydantic_settings import BaseSettings, SettingsConfigDict

from pi_conf import Config, load_config
from pi_conf.config import load_from_appname


sentinel = object()


class ConfigDict(SettingsConfigDict, total=False):
    """Extended SettingsConfigDict for TOML-specific settings.

    Attributes:
        appname (str): The name of the application.
        toml_table_header (str): Header for the TOML table.
    """

    appname: str
    toml_table_header: str


class ConfigSettings(BaseSettings):
    """Main class for handling configuration settings.

    Attributes:
        model_config (ConfigDict): Configuration for the model.
    """

    model_config = ConfigDict(toml_table_header="")

    def __init__(self, *args, **kwargs):
        """Initialize the ConfigSettings object.

        Args:
            *args: Variable length argument list.
            **kwargs: Arbitrary keyword arguments.

        Raises:
            ValueError: If neither 'toml_file' nor 'appname' is provided.
        """
        model_config = kwargs.pop("model_config", self.model_config)

        specified_vars = ["toml_file", "appname", "toml_table_header"]
        for var in specified_vars:
            val = kwargs.pop(var, sentinel)
            if val is not sentinel:
                model_config[var] = val

        cfg = None
        appname = model_config.get("appname")
        toml_file = model_config.get("toml_file")
        if toml_file and appname:
            ## we want to search for the toml file in the appname directory
            cfg = load_from_appname(appname=appname, file=toml_file)
        if not cfg:
            locs = ["toml_file", "appname"]
            for loc in locs:
                v = model_config.get(loc)
                if v:
                    cfg = load_config(v)
                    break
            else:
                raise ValueError(f"one of {locs} must be provided")

        toml_table_header = model_config.get("toml_table_header", "")
        if toml_table_header:
            try:
                cfg = cfg.get_nested(toml_table_header)
            except KeyError:
                raise KeyError(
                    f"toml_table_header '{toml_table_header}' not "
                    f"found in config {cfg.provenance[-1].source}"
                )

        temp_toml_file = self.create_temp_toml_file(cfg)
        try:
            self.parse_nested_objects(cfg)
            super().__init__(*args, **cfg, **kwargs)
        finally:
            import os

            os.unlink(temp_toml_file)

    @classmethod
    def model_construct(cls, _fields_set: set[str] | None = None, **values: Any) -> Self:
        """Construct the model.

        Args:
            _fields_set: Set of fields to be set (not currently supported).
            **values: Arbitrary keyword arguments for model construction.

        Returns:
            Self: An instance of the class.

        Raises:
            NotImplementedError: If _fields_set is provided.
        """
        if _fields_set is not None:
            raise NotImplementedError(
                "ConfigSettings.model_construct does not currently support _fields_set"
            )

        return cls(**values)

    @classmethod
    def create_temp_toml_file(cls, config: Config) -> str:
        """Create a temporary TOML file from the given configuration.

        Args:
            config (Config): The configuration object.

        Returns:
            str: The path to the created temporary TOML file.
        """
        with tempfile.NamedTemporaryFile(mode="w+", suffix=".toml", delete=False) as temp_file:
            toml.dump(config, temp_file)
            return temp_file.name

    def parse_nested_objects(self, config_dict: Dict[str, Any]):
        """Parse nested objects in the configuration dictionary.

        Args:
            config_dict (Dict[str, Any]): The configuration dictionary to parse.
        """
        for field_name, field_value in self.__annotations__.items():
            if field_name in config_dict:
                config_dict[field_name] = self._parse_field(field_value, config_dict[field_name])

    def _parse_field(self, field_type, field_value):
        """Parse a single field based on its type.

        Args:
            field_type: The type of the field.
            field_value: The value of the field.

        Returns:
            The parsed field value.
        """
        origin = get_origin(field_type)
        if origin is List:
            item_type = get_args(field_type)[0]
            if isinstance(item_type, type) and issubclass(item_type, BaseModel):
                return [self._parse_model(item_type, item) for item in field_value]
        elif isinstance(field_type, type) and issubclass(field_type, BaseModel):
            return self._parse_model(field_type, field_value)
        return field_value

    def _parse_model(self, model_class: Type[BaseModel], data: Dict[str, Any]) -> BaseModel:
        """Parse and validate a model using Pydantic.

        Args:
            model_class (Type[BaseModel]): The Pydantic model class.
            data (Dict[str, Any]): The data to parse into the model.

        Returns:
            BaseModel: An instance of the parsed model.

        Raises:
            ValidationError: If the data fails to validate against the model.
        """
        try:
            return model_class(**data)
        except ValidationError as e:
            print(f"Error parsing {model_class.__name__}: {e}")
            raise e

    def _pformat(self, indent: int = 4) -> str:
        return json.dumps(self.model_dump(mode="json"), indent=indent)



from pi_conf.config import (
    AttrDict,
    Config,
    ProvenanceDict,
    cfg,
    load_config,
    set_config,
)

__all__ = [
    "load_config",
    "set_config",
    "cfg",
    "Config",
    "AttrDict",
    "ProvenanceDict",
]

try: ## Optional pydantic settings support
    from pi_conf.config_settings import ConfigSettings, ConfigDict

    __all__.extend(["ConfigSettings", "ConfigDict"])
except ImportError:
    pass


"""Config"""

import configparser
import json
import logging
import os
from typing import Optional, TypeVar

from pi_conf.attr_dict import AttrDict, has_yaml, is_tomllib, site_config_dir
from pi_conf.module_check import has_yaml, is_tomllib
from pi_conf.provenance import Provenance, ProvenanceOp
from pi_conf.provenance import get_provenance_manager as get_pmanager

if has_yaml:
    import yaml
if is_tomllib:
    import tomllib
else:
    import toml

T = TypeVar("T", bound="AttrDict")

log = logging.getLogger(__name__)


class ProvenanceDict(AttrDict):
    """Config class, an attr dict that allows referencing by attribute and also
    tracks provenance information, such as updates and where they were from.
    Example:
        cfg = Config({"a":1, "b":{"c":3}})
        cfg.a.b.c == cfg["a"]["b"]["c"] # True
    """

    def __init__(self, *args, **kwargs):
        enable_provenance = kwargs.pop("enable_provenance", True)
        get_pmanager().set_enabled(self, enable_provenance)

        super().__init__(*args, **kwargs)
        self.__dict__ = self
        get_pmanager().append(self, Provenance("dict", ProvenanceOp.set))

    def __post_init__(self):
        ## Iterate over members and add them to the dict
        members = [
            attr
            for attr in dir(self)
            if not callable(getattr(self, attr)) and not attr.startswith("__")
        ]
        for m in members:
            if m == "provenance":
                continue
            if isinstance(getattr(self, m), dict):
                nd = AttrDict.from_dict(getattr(self, m))
                setattr(self, m, nd)
                self[m] = nd
            else:
                self[m] = getattr(self, m)

    def load_config(
        self,
        appname_path_dict: str | dict,
        directories: Optional[str | list] = None,
    ) -> None:
        """Loads a config based on the given appname | path | dict

        Args:
            appname_path_dict (str): Set the config from an appname | path | dict
            Can be passed with the following.
                Dict: updates cfg with the given dict
                str: a path to a (.toml|.json|.ini|.yaml) file
                str: appname to search for the config.toml in the the application config dir
            directories (Optional[str | list]): Optional list of directories to search
        """
        newcfg = load_config(appname_path_dict, directories=directories)
        self.update(newcfg, _add_to_provenance=False)
        get_pmanager().extend(self, newcfg.provenance)
        get_pmanager().delete(newcfg)

    @property
    def provenance(self) -> list[Provenance]:
        return get_pmanager().get(self)

    def __del__(self):
        """Delete the config from the provenance if this object is deleted"""
        get_pmanager().delete(self)

    def update(self, *args, **kwargs):
        """Update the config with another dict"""
        _add_to_provenance = kwargs.pop("_add_to_provenance", True)
        super().update(*args, **kwargs)
        if _add_to_provenance:
            get_pmanager().append(self, Provenance("dict", ProvenanceOp.update))

    def clear(self) -> None:
        get_pmanager().clear(cfg)
        return super().clear()

    @classmethod
    def from_dict(cls: type[T], d: dict, _nested_same_class: bool = False) -> T:
        """Make an AttrDict object without any keys
        that will overwrite the normal functions of a dict

        Args:
            cls (Type[AttrDict]): Create a new AttrDict object (or subclass)
            d (dict): The dictionary to convert to an AttrDict
            _nested_same_class (bool): If True, nested dicts will be the subclass,
                else they will be AttrDict

        Returns:
            AttrDict: the AttrDict object, or subclass
        """
        ad = cls._from_dict(d, _nested_same_class=_nested_same_class, _depth=0)
        return ad


class Config(ProvenanceDict):
    pass


def _load_config_file(path: str, ext: Optional[str] = None) -> Config:
    """Load a config file from the given path"""
    if ext is None:
        __, ext = os.path.splitext(path)

    if ext == ".toml":
        if is_tomllib:  ## python 3.11+ have toml in the core libraries
            with open(path, "rb") as fp:
                return Config.from_dict(tomllib.load(fp))  # type: ignore
        else:  ## python <3.11 need the toml library
            return Config.from_dict(toml.load(path))  # type: ignore
    elif ext == ".json":
        with open(path, "r") as fp:
            return Config.from_dict(json.load(fp))
    elif ext == ".ini":
        cfg_parser = configparser.ConfigParser()
        cfg_parser.read(path)
        cfg_dict = {section: dict(cfg_parser[section]) for section in cfg_parser.sections()}
        return Config.from_dict(cfg_dict)
    elif ext == ".yaml":
        if not has_yaml:
            raise Exception(
                "Error! YAML not installed. If you would like to use YAML with pi-conf, "
                "install it with 'pip install pyyaml' or 'pip install pi-conf[yaml]"
            )
        with open(path, "r") as fp:
            return Config.from_dict(yaml.safe_load(fp))  # type: ignore
    raise Exception(f"Error! Unknown config file extension '{ext}'")


def _get_default_search_paths(filename: str, appname: Optional[str] = None) -> list[str]:
    """Get the default search paths for a config file."""
    if appname:
        return [
            os.path.expanduser(f"~/.config/{appname}/{filename}"),
            os.path.join(site_config_dir(appname=appname), filename),
        ]
    else:
        return [
            filename,
            os.path.expanduser(f"~/.config/{filename}"),
            os.path.join(site_config_dir(), filename),
        ]


def _get_search_paths(
    filename: str, directories: Optional[str | list[str]], appname: Optional[str] = None
) -> list[str]:
    """Get the search paths based on provided directories or default locations."""
    if directories is None:
        return _get_default_search_paths(filename, appname)
    elif isinstance(directories, str):
        return [os.path.join(os.path.expanduser(directories), filename)]
    else:
        return [os.path.join(os.path.expanduser(d), filename) for d in directories]


def _find_file_with_extensions(path: str, extensions: list[str]) -> Optional[str]:
    """Find a file with given extensions."""
    if os.path.exists(path):
        return path
    for ext in extensions:
        full_path = path.replace("<ext>", ext)
        full_path = os.path.expanduser(full_path)
        if os.path.isfile(full_path):
            log.debug(f"Found config: '{full_path}'")
            return full_path
    return None


def _find_config(
    config_file_or_appname: str, directories: Optional[str | list[str]] = None
) -> Optional[str]:
    """Find the config file from the config directory or direct path."""
    # First, check if it's a direct file path
    if os.path.isfile(config_file_or_appname):
        return config_file_or_appname

    # If not a direct file path, check if it looks like a filename (has an extension)
    _, ext = os.path.splitext(config_file_or_appname)
    if ext:
        # It looks like a filename
        search_paths = _get_search_paths(config_file_or_appname, directories=directories)
    else:
        # It looks like an appname
        search_paths = _get_search_paths(
            "config.<ext>", directories=directories, appname=config_file_or_appname
        )

    extensions = ["toml", "json", "ini", "yaml"] if not ext else [""]
    for path in search_paths:
        found_path = _find_file_with_extensions(path, extensions)
        if found_path:
            log.debug(f"Found config: '{found_path}'")
            return found_path

    return None


def _find_config_from_appname(
    appname: str, file: Optional[str] = None, directories: Optional[str | list[str]] = None
) -> Optional[str]:
    """
    Find a config file based on the appname and optionally a specific file name.
    """
    filename = file or "config.<ext>"
    search_paths = _get_search_paths(filename, directories=directories, appname=appname)
    extensions = ["toml", "json", "ini", "yaml"] if not file else [""]

    for path in search_paths:
        found_path = _find_file_with_extensions(path, extensions)
        if found_path:
            log.debug(f"Found config: '{found_path}'")
            return found_path

    return None


def update_config(appname_path_dict: str | dict, directories: Optional[str | list[str]]) -> Config:
    """Update the global config with another config

    Args:
        appname_path_dict (str): Set the config from an appname | path | dict
        Can be passed with the following.
            Dict: updates cfg with the given dict
            str: a path to a (.toml|.json|.ini|.yaml) file
            str: appname to search for the config.toml in the the application config dir

    Returns:
        Config: A config object (an attribute dictionary)
    """
    newcfg = load_config(appname_path_dict, directories=directories)
    cfg.update(newcfg, _add_to_provenance=False)
    get_pmanager().extend(cfg, newcfg.provenance)
    get_pmanager().delete(newcfg)
    return cfg


def set_config(
    appname_path_dict: Optional[str | dict] = None,
    create_if_not_exists: bool = True,
    create_with_extension=".toml",
    directories: Optional[str | list] = None,
) -> Config:
    """Sets the global config.toml to use based on the given appname | path | dict

    Args:
        appname_path_dict (str): Set the config from an appname | path | dict
            Can be passed with the following.
                Dict: updates cfg with the given dict
                str: a path to a (.toml|.json|.ini|.yaml) file
                str: appname to search for the config.toml in the the application config dir
        create_if_not_exists (bool): If True, and appname_path_dict is a path, create the config file if it doesn't exist
        create_with_extension (str): The extension to use if creating the config file
        directories (Optional[str | list]): Optional list of directories to search

    Returns:
        Config: A config object (an attribute dictionary)
    """
    if appname_path_dict is None:
        appname_path_dict = ".config.toml"

    cfg.clear()
    if create_if_not_exists and isinstance(appname_path_dict, str):
        path = _find_config(appname_path_dict, directories=directories)
        if path is None:
            if not isinstance(appname_path_dict, str):
                raise Exception("Error! appname_path_dict must be a string to create a config file")

            if directories is not None:
                if isinstance(directories, str):
                    directories = [directories]
                path = directories[0]
            else:
                path = site_config_dir(appname=appname_path_dict)
            path = os.path.join(path, appname_path_dict, f"config{create_with_extension}")
            os.makedirs(os.path.dirname(path), exist_ok=True)

            log.info(f"Creating config file at '{path}' for appname '{appname_path_dict}'")
            with open(path, "w") as fp:
                fp.write("")
            directories = [os.path.dirname(path)]
    return update_config(appname_path_dict, directories=directories)


def load_from_dict(d: dict) -> Config:
    """Load a config from a dict"""
    return Config.from_dict(d)


def load_from_path(path: str, directories: Optional[str | list] = None) -> Config:
    """Load a config from a file path"""
    full_path = _find_config(path, directories=directories)
    if full_path is None:
        raise FileNotFoundError(f"No config file found at '{path}' or in provided directories")
    newcfg = _load_config_file(full_path)
    get_pmanager().set(newcfg, Provenance(full_path, ProvenanceOp.set))
    return newcfg


def load_from_appname(
    appname: str, file: Optional[str] = None, directories: Optional[str | list[str]] = None
) -> Config:
    """
    Load a config from an appname, optionally specifying a file name.

    Args:
        appname (str): The name of the application
        file (Optional[str]): Specific file to search for. If None, defaults to 'config.<ext>'
        directories (Optional[str | list[str]]): Optional list of directories to search

    Returns:
        Config: A config object (an attribute dictionary)

    Raises:
        FileNotFoundError: If no config file is found
    """
    config_path = _find_config_from_appname(appname, file, directories)

    if config_path is None:
        filestr = f" with file '{file}'" if file else ""
        log.warning(f"No config file found for appname '{appname}' {filestr}")
        log.warning(f"You can create a config file at '{site_config_dir(appname=appname)}' {filestr}")
        raise FileNotFoundError(f"No config file found for '{appname}' {filestr}")

    return load_from_path(config_path)


def load_config(
    appname_path_dict: Optional[str | dict] = None,
    file: Optional[str] = None,
    directories: Optional[str | list] = None,
    ignore_warnings: bool = False,
) -> Config:
    """Loads a config based on the given appname | path | dict

    Args:
        appname_path_dict (str | dict): Set the config from an appname | path | dict
        Can be passed with the following:
            Dict: updates cfg with the given dict
            str: a path to a (.toml|.json|.ini|.yaml) file
            str: appname to search for the config.toml in the application config dir
        file (Optional[str]): Specific file to search for when appname is provided
        directories (Optional[str | list]): Optional list of directories to search
        ignore_warnings (bool): If True, suppress warnings and return an empty Config for missing files

    Returns:
        Config: A config object (an attribute dictionary)
    """
    if appname_path_dict is None:
        appname_path_dict = ".config.toml"

    if isinstance(appname_path_dict, dict):
        return load_from_dict(appname_path_dict)

    try:
        return load_from_path(appname_path_dict, directories)
    except FileNotFoundError:
        # If it's not found as a direct path, try as an appname
        try:
            return load_from_appname(appname_path_dict, file, directories)
        except FileNotFoundError:
            if ignore_warnings:
                return Config.from_dict({})
            raise


cfg = Config()  ## Our global config



import importlib.util


def check_module(module_name):
    return importlib.util.find_spec(module_name) is not None


has_yaml = check_module("yaml")
is_tomllib = check_module("tomllib")

if not is_tomllib:
    is_tomllib = check_module("toml")



import inspect
import logging
import os
from collections import defaultdict
from dataclasses import dataclass, field
from enum import Enum
from typing import Optional

log = logging.getLogger(__name__)

PROVENANCE_DEPTH = 2


class ProvenanceOp(str, Enum):
    set = "set"
    update = "update"
    clear = "clear"
    clear_and_set = "clear_and_set"

    def __str__(self):
        return self.value


class Provenance:
    """Provenance of the config"""

    def __init__(self, source: str, operation: str, stack: Optional[list[str]] = None):
        self.source = source
        self.stack = stack
        self.operation = operation
        if self.stack is None:
            self.stack = _provenance_manager.get_methods_that_called_this_method(PROVENANCE_DEPTH)

    def __repr__(self):
        return (
            f"<Provenance: abbr_stack='{self.stack}' source='{self.source}' op='{self.operation}'>"
        )

    def __str__(self):
        return (
            f"<Provenance: abbr_stack='{self.stack}' source='{self.source}'> op='{self.operation}'"
        )


@dataclass
class ProvenanceManager:
    """Provenance manager"""

    _provenance: dict[int, list[Provenance]] = field(default_factory=lambda: defaultdict(list))
    _enabled: set[int] = field(default_factory=set)

    def set_enabled(self, obj, enable: bool = True):
        """Set whether or not to enable provenance"""
        if enable:
            self._enabled.add(id(obj))
        else:
            self._enabled.discard(id(obj))

    def get(self, obj) -> list[Provenance]:
        """Get the provenance of the given object"""
        return self._provenance.get(id(obj), [])

    def set(self, obj, provenance: Provenance | list[Provenance]):
        """Set the provenance of the given object"""
        oid = id(obj)
        if oid not in self._enabled:
            return
        if isinstance(provenance, Provenance):
            provenance = [provenance]
        self._provenance[oid] = provenance

    def append(self, obj, provenance: Provenance):
        """Append to the provenance of the given object"""
        oid = id(obj)
        if oid not in self._enabled:
            return
        self._provenance[oid].append(provenance)

    def extend(self, obj, provenance: list[Provenance]):
        """Extend the provenance of the given object"""
        oid = id(obj)
        if oid not in self._enabled:
            return
        self._provenance[id(obj)].extend(provenance)

    def clear(self, obj):
        """Clear the provenance of the given object"""
        oid = id(obj)
        if oid not in self._enabled:
            return
        self._provenance[id(obj)] = []

    def __repr__(self):
        return f"<ProvenanceManager: {self._provenance}>"

    def delete(self, obj):
        """Delete the provenance of the given object"""
        try:
            del self._provenance[id(obj)]
        except KeyError:
            pass

    @staticmethod
    def get_methods_that_called_this_method(depth: Optional[int] = None) -> list[str]:
        """Get the method that called this method"""
        try:
            stack = []
            for i in range(len(inspect.stack())):
                frame = inspect.stack()[i]
                lineno = frame.lineno
                module_path = frame.filename
                base_name = os.path.basename(module_path)
                n = os.path.basename(os.path.dirname(module_path))
                if n == "pi_conf":
                    continue
                fn = frame.function
                stack.append(f"{base_name}::{fn}:{lineno}")
            if depth is not None:
                return stack[-depth:][::-1]
            return stack[::-1]

        except Exception as e:
            log.error(f"Error! {e}")
            return ["Unknown"]


@dataclass
class NullOpProvenanceManager(ProvenanceManager):
    """Null op provenance manager"""

    def set_enabled(self, obj, enable: bool = True):
        pass

    def get(self, obj) -> list[Provenance]:
        """Get the provenance of the given object"""
        return []

    def set(self, obj, provenance: Provenance | list[Provenance]):
        """Set the provenance of the given object"""

    def append(self, obj, provenance: Provenance):
        """Append to the provenance of the given object"""

    def extend(self, obj, provenance: list[Provenance]):
        """Extend the provenance of the given object"""

    def clear(self, obj):
        """Clear the provenance of the given object"""

    def __repr__(self):
        return f"<NullOpProvenanceManager>"

    def delete(self, obj):
        """Delete the provenance of the given object"""

    @staticmethod
    def get_methods_that_called_this_method(depth: Optional[int] = None) -> list[str]:
        """Get the method that called this method"""
        return ["Unknown"]


_provenance_manager = ProvenanceManager()  ## provenance of the config


def set_use_provenance(use_provenance: bool = True):
    """Set whether or not to use provenance"""
    global _provenance_manager
    if use_provenance:
        _provenance_manager = ProvenanceManager()
    else:
        _provenance_manager = NullOpProvenanceManager()


def get_provenance_manager() -> ProvenanceManager:
    """Get the provenance manager"""
    return _provenance_manager
